<?php

namespace Drupal\tapis_job\Controller;

use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Url;
use Drupal\tapis_app\DrupalIds as AppDrupalIds;
use Drupal\tapis_job\TapisProvider\TapisJobProviderInterface;
use Drupal\tapis_tenant\TapisProvider\TapisSiteTenantProviderInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\RedirectResponse;
use Symfony\Component\HttpFoundation\Request;

/**
 *
 */
class JobCancelController extends ControllerBase {

  /**
   * The Tapis site tenant provider.
   *
   * @var \Drupal\tapis_tenant\TapisProvider\TapisSiteTenantProviderInterface
   */
  protected TapisSiteTenantProviderInterface $tapisSiteTenantProvider;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The Tapis job provider.
   *
   * @var \Drupal\tapis_job\TapisProvider\TapisJobProviderInterface
   */
  protected TapisJobProviderInterface $tapisJobProvider;

  /**
   * Constructs a new JobCancelController object.
   *
   * @param \Drupal\tapis_tenant\TapisProvider\TapisSiteTenantProviderInterface $tapisSiteTenantProvider
   *   The Tapis site tenant provider.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager.
   * @param \Drupal\tapis_job\TapisProvider\TapisJobProviderInterface $tapisJobProvider
   *   The Tapis job provider.
   */
  public function __construct(TapisSiteTenantProviderInterface $tapisSiteTenantProvider,
                              EntityTypeManagerInterface $entityTypeManager,
                              TapisJobProviderInterface $tapisJobProvider) {
    $this->tapisSiteTenantProvider = $tapisSiteTenantProvider;
    $this->entityTypeManager = $entityTypeManager;
    $this->tapisJobProvider = $tapisJobProvider;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get("tapis_tenant.tapis_site_tenant_provider"),
      $container->get('entity_type.manager'),
      $container->get('tapis_job.tapis_job_provider')
    );
  }

  /**
   * Terminate a job and redirects to the job view page.
   *
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The current request object.
   * @param int $tapis_job
   *   The job ID.
   *
   * @return \Symfony\Component\HttpFoundation\RedirectResponse
   *   A redirect response to the job view page.
   */
  public function cancelJob(Request $request, int $tapis_job) {
    // Load the job entity.
    /** @var \Drupal\tapis_job\TapisJobInterface $job */
    $job = $this->entityTypeManager->getStorage('tapis_job')->load($tapis_job);

    if (!$job) {
      $this->messenger()->addError($this->t('The job does not exist.'));
      return $this->redirectJobViewPage($tapis_job);
    }

    $tenantId = $job->getTenantId();
    $jobOwnerId = $job->getOwnerId();

    // Check if the tenant is in maintenance mode.
    if ($this->tapisSiteTenantProvider->isTenantInMaintenanceMode($tenantId)) {
      $this->messenger()
        ->addError($this->t('This job cannot be terminated because its site is in maintenance mode.'));
      return $this->redirectJobViewPage($tapis_job);
    }

    // Cancel the job using the job provider.
    $this->tapisJobProvider->cancelJob($tenantId, $job->getTapisUUID(), $jobOwnerId);

    $app = $job->getApp();
    $appType = $app->get(AppDrupalIds::APP_TYPE)->getValue()[0]['value'] ?? '';
    if (in_array($appType, ["web", "vnc"])) {
      $this->tapisJobProvider->deleteAllAccessLinksForJob($job->id());
    }

    $jobResult = $this->tapisJobProvider->getJob($tenantId, $job->getTapisUUID(), $jobOwnerId);
    $jobStatus = $jobResult['status'];

    $this->messenger()
      ->addStatus(($this->t('The job was successfully terminated; however, it may take some time to update the job status.')));
    return $this->redirectJobViewPage($tapis_job);
  }

  /**
   * Redirects to the job view page.
   *
   * @param int $tapis_job_id
   *   The job ID.
   *
   * @return \Symfony\Component\HttpFoundation\RedirectResponse
   *   A redirect response to the job view page.
   */
  private function redirectJobViewPage(int $tapis_job_id): RedirectResponse {
    return new RedirectResponse(Url::fromRoute('entity.tapis_job.canonical', ['tapis_job' => $tapis_job_id])->toString());
  }

}
